from datetime import datetime
from enum import IntEnum
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from fastapi import APIRouter, Body, Depends, Path, Query, status
from fastapi.responses import ORJSONResponse
from pydantic import BaseModel as PydanticBaseModel
from pydantic import create_model
from pypika.terms import Term
from tortoise.fields.relational import ForeignKeyFieldInstance, RelationalField
from tortoise.functions import Function, Max
from tortoise.models import Field, MetaInfo
from tortoise.queryset import Q, QuerySet
from tortoise.transactions import in_transaction

from .. import json
from ..model import BaseModel
from .exceptions import HTTPNotFoundError, HTTPPreconditionRequiredError

UserModel = TypeVar("UserModel")
Model = TypeVar("Model", bound=BaseModel)
CreateForm = TypeVar("CreateForm", bound=PydanticBaseModel)
UpdateForm = TypeVar("UpdateForm", bound=PydanticBaseModel)


class ResourceMethod(IntEnum):
    list = 0
    get = 1
    create = 2
    put = 3
    delete = 4


class DeleteResponse(PydanticBaseModel):
    id: Optional[Union[str, int]]


class API(Generic[UserModel, Model, CreateForm, UpdateForm]):
    def __init__(
        self,
        model: Type[Model],
        create_form: Type[CreateForm],
        update_form: Type[UpdateForm],
        get_current_user: Callable[
            ..., Union[Coroutine[Any, Any, UserModel], UserModel]
        ],
        enable_create: bool = True,
        enable_update: bool = True,
        enable_delete: bool = True,
        enable_get: bool = True,
    ) -> None:
        super().__init__()
        self.model = model
        self.create_form = create_form
        self.update_form = update_form
        self.get_current_user = get_current_user
        self.enable_create = enable_create
        self.enable_update = enable_update
        self.enable_delete = enable_delete
        self.enable_get = enable_get

    def translate_id(self, user: UserModel, id: str) -> str:
        return id

    def translate_condition(
        self, user: UserModel, q: QuerySet[Model], k: str, v: Any
    ) -> Any:
        return v

    def group_by_f(self, group_by: str, field: Field) -> Union[Function, Term]:
        return Max(group_by)

    def group_by(
        self, user: UserModel, model: Type[Model], group_by: List[str]
    ) -> QuerySet[Model]:
        if group_by:
            model_meta: MetaInfo = getattr(model, "_meta")
            field_names: Iterable[str] = model_meta.fields_map.keys()
            annotates: Dict[str, Union[Function, Term]] = {}
            for field_name in field_names:
                if field_name in group_by:
                    continue
                field = model_meta.fields_map[field_name]
                if isinstance(field, ForeignKeyFieldInstance):
                    field_name = f"${field_name}_id"
                elif isinstance(field, RelationalField):
                    continue
                annotates[field_name] = self.group_by_f(field_name, field)
            q = model.annotate(**annotates)
            q = q.group_by(*group_by)
            return q
        return model.all()

    def q(
        self,
        user: UserModel,
        q: QuerySet[Model],
        method: ResourceMethod = ResourceMethod.list,
    ) -> QuerySet[Model]:
        return q

    async def before_create(
        self, user: UserModel, obj: Model, input: CreateForm
    ) -> Model:
        return obj

    async def after_create(self, user: UserModel, obj: Model, input: CreateForm) -> Any:
        return obj

    async def before_update(
        self, user: UserModel, obj: Model, input: UpdateForm
    ) -> Model:
        return obj

    async def after_update(self, user: UserModel, obj: Model, input: UpdateForm) -> Any:
        return obj

    async def before_delete(self, user: UserModel, obj: Model) -> Model:
        return obj

    @classmethod
    def ids_path(cls):
        return Path(..., title="ID", description="支持采用 `1,2,3` 形式传入多个")

    @classmethod
    def include_query(cls):
        return Query(
            [],
            title="只获取某些字段",
            description="支持采用 `include=id&include=name` 形式传入多个",
        )

    @classmethod
    def prefetch_query(cls):
        return Query(
            [],
            title="动态加载计算字段",
            description="获取更多字段, 支持采用 `prefetch=id&prefetch=user.id` 形式传入多个",
        )

    @classmethod
    def group_by_query(cls):
        return Query(
            [],
            title="分组",
            description="分组, 不在分组字段默认为 MAX, 支持采用 `group_by=x&group_by=y` 形式传入多个",
        )

    @classmethod
    def get_form_type(cls, form: Any) -> Any:
        f = getattr(form, "form_type", None)
        return f and f() or Body(...)

    @property
    def connection_name(self):
        return self.model._meta.default_connection  # type: ignore

    def bind(self, router: APIRouter):
        ListPydanticModel = self.model.list()
        DetailPydanticModel = self.model.detail()
        CreateForm = self.create_form
        UpdateForm = self.update_form

        DetailPydanticModels = Union[
            DetailPydanticModel, List[DetailPydanticModel]  # type: ignore
        ]
        Response = create_model(
            f"{self.model.__name__}.Response",
            __base__=PydanticBaseModel,
            total=0,
            data=(List[ListPydanticModel], ...),  # type: ignore
        )

        methods: Dict[str, Callable[..., Any]] = {}

        if self.enable_create:

            @router.post(
                "/",
                response_model=DetailPydanticModels,
                response_class=ORJSONResponse,
                status_code=status.HTTP_201_CREATED,
            )
            async def create(
                input: Union[CreateForm, List[CreateForm]] = self.get_form_type(
                    CreateForm
                ),
                prefetch: List[str] = self.prefetch_query(),
                current_user: UserModel = Depends(self.get_current_user),
            ) -> Any:
                async with in_transaction(self.connection_name):
                    is_batch = isinstance(input, list)
                    inputs = cast(List[CreateForm], input if is_batch else [input])

                    returns: List[PydanticBaseModel] = []
                    for inpu in inputs:
                        raw: Dict[str, Any] = inpu.dict(exclude_unset=True)
                        raw, m2m_raw = await self.model.save_related(raw)
                        obj = self.model(**raw)

                        obj_rtn = await self.before_create(current_user, obj, inpu)
                        if obj_rtn is not obj:
                            obj = obj_rtn

                        await obj.save()
                        prefetch = list(set([*prefetch, *m2m_raw.keys()]))
                        for k, v in m2m_raw.items():
                            await getattr(obj, k).add(*v)
                        if prefetch:
                            await obj.fetch_related(*prefetch)

                        obj_rtn = await self.after_create(current_user, obj, inpu)
                        if obj_rtn and obj_rtn is not obj:
                            obj = obj_rtn

                        if isinstance(obj, PydanticBaseModel):
                            returns.append(obj)
                        else:
                            returns.append(DetailPydanticModel.from_orm(obj))
                    return returns if is_batch else returns[0]

            methods["create"] = create

        if self.enable_get:
            help = """支持采用 `condition=...&condition=...` 传入多个, 使用 JSON 序列化
[查询语法参考](https://tortoise-orm.readthedocs.io/en/latest/query.html#filtering)

###### 简单查询

```javascript
{
    "email": "name@example.com",
    "name.contains": "name",
    "id.in": [1, 2, 3],
}
```

###### 跨表查询

```javascript
{
    "user.email": "name@example.com",
    "user.name.contains": "name",
}
```

###### 子逻辑, 可选项: `.and` `.or` `.not` `.not_or`

```javascript
{
    ".or": {
        "name.contains": "name2",
        "email.contains": "example2.com",
    },
    "name.contains": "name",
    "email.contains": "example.com",
}
```

###### 运算方式, 可选项: `and` `or` `not` `not_or`

```javascript
{
    ".logic": "or",
    "name.contains": "name",
    "email.contains": "example.com",
}
```
            """

            @router.get("/", response_model=Response, response_class=ORJSONResponse)
            async def index(
                offset: int = Query(0, title="分页偏移"),
                limit: int = Query(20, title="分页限额"),
                condition: List[str] = Query([], title="查询条件", description=help),
                order_by: List[str] = Query(
                    [],
                    title="排序",
                    description="支持采用 `order_by=id&order_by=user.id` 形式传入多个",
                ),
                include: List[str] = self.include_query(),
                prefetch: List[str] = self.prefetch_query(),
                group_by: List[str] = self.group_by_query(),
                current_user: UserModel = Depends(self.get_current_user),
            ) -> Any:
                q = self.group_by(current_user, self.model, group_by)
                q = self.q(current_user, q, ResourceMethod.list)

                # 通用过滤方案
                # https://tortoise-orm.readthedocs.io/en/latest/query.html
                if condition:
                    for raw in condition:
                        kv = json.loads(raw)
                        reverse = False
                        join_type = kv.pop(".logic", "AND").upper()
                        if join_type == "NOT":
                            reverse = True
                            join_type = "AND"
                        elif join_type == "NOT_OR":
                            reverse = True
                            join_type = "OR"
                        q_kwargs: Dict[str, Any] = {}
                        for k, v in kv.items():
                            k = self.model.normalize_field(k)
                            v = self.translate_condition(current_user, q, k, v)
                            if isinstance(v, QuerySet):
                                q = cast(QuerySet[Model], v)
                            elif v is not None:
                                q_kwargs[k] = v
                        condition_q = Q(
                            **q_kwargs,
                            join_type=join_type,
                        )
                        if reverse:
                            condition_q = ~condition_q
                        q = q.filter(condition_q)

                if include:
                    q = q.only(*include)

                q = q.distinct()
                total = await q.count()

                if order_by:
                    for item in order_by:
                        q = q.order_by(self.model.normalize_field(item))

                q = q.offset(offset).limit(limit)
                if group_by:
                    dicts = cast(List[Dict[str, Any]], await q.values())
                    return Response(
                        total=total,
                        data=[ListPydanticModel(**d) for d in dicts],
                    )

                objs = await q
                if prefetch:
                    for obj in objs:
                        await obj.fetch_related(*prefetch)

                return Response(
                    total=total,
                    data=[ListPydanticModel.from_orm(obj) for obj in objs],
                )

            @router.get(
                "/{id}",
                response_model=DetailPydanticModel,
                response_class=ORJSONResponse,
            )
            async def get(
                id: str = Path(..., title="id"),
                prefetch: List[str] = self.prefetch_query(),
                current_user: UserModel = Depends(self.get_current_user),
            ) -> Any:
                id = self.translate_id(current_user, id)
                q = self.q(
                    current_user,
                    self.model.all().filter(id=id),
                    ResourceMethod.get,
                )
                obj = await q.first()
                if not obj:
                    raise HTTPNotFoundError
                if prefetch:
                    await obj.fetch_related(*prefetch)
                return DetailPydanticModel.from_orm(obj)

            methods["index"] = index
            methods["get"] = get

        if self.enable_update:

            @router.put(
                "/{ids}",
                response_model=DetailPydanticModels,
                response_class=ORJSONResponse,
            )
            async def update(
                ids: str = self.ids_path(),
                input: UpdateForm = self.get_form_type(UpdateForm),
                prefetch: List[str] = self.prefetch_query(),
                current_user: UserModel = Depends(self.get_current_user),
            ) -> Any:
                async with in_transaction(self.connection_name):
                    rtns: List[Any] = []
                    raw = input.dict(exclude_unset=True)
                    for id in ids.split(","):
                        id = id.strip()
                        id = self.translate_id(current_user, id)
                        obj = await self.q(
                            current_user,
                            self.model.filter(id=id),
                            ResourceMethod.put,
                        ).get()

                        obj_rtn = await self.before_update(current_user, obj, input)
                        if obj_rtn is not obj:
                            obj = obj_rtn

                        updated_at = raw.pop("updated_at", None)
                        if updated_at:
                            # 防止老数据修改
                            if isinstance(updated_at, str):
                                updated_at = json.parse_iso_datetime(updated_at)
                            if isinstance(updated_at, datetime):
                                if obj.updated_at > updated_at:
                                    raise HTTPPreconditionRequiredError

                        raw, m2m_raw = await self.model.save_related(raw)

                        update_fields = raw.keys()
                        if update_fields:
                            obj.update(raw)
                            await obj.save(update_fields=update_fields)

                        obj_prefetch = list(set([*prefetch, *m2m_raw.keys()]))
                        for k, v in m2m_raw.items():
                            await getattr(obj, k).add(*v)
                        if obj_prefetch:
                            await obj.fetch_related(*obj_prefetch)

                        obj_rtn = await self.after_update(current_user, obj, input)
                        if obj_rtn and obj_rtn is not obj:
                            obj = obj_rtn

                        rtns.append(
                            obj
                            if isinstance(obj, PydanticBaseModel)
                            else DetailPydanticModel.from_orm(obj)
                        )
                    return len(rtns) > 1 and rtns or rtns[0]

            methods["update"] = update

        if self.enable_delete:

            @router.delete(
                "/{ids}",
                response_model=Union[DeleteResponse, List[DeleteResponse]],
                response_class=ORJSONResponse,
            )
            async def delete(
                ids: str = self.ids_path(),
                current_user: UserModel = Depends(self.get_current_user),
            ) -> Union[DeleteResponse, List[DeleteResponse]]:
                async with in_transaction(self.connection_name):
                    rs: List[DeleteResponse] = []
                    for id in ids.split(","):
                        id = self.translate_id(current_user, id)
                        obj = await self.q(
                            current_user,
                            self.model.filter(id=id),
                            ResourceMethod.delete,
                        ).get()
                        obj_rtn = await self.before_delete(current_user, obj)
                        if obj_rtn is not obj:
                            obj = obj_rtn
                        await obj.delete()
                        rs.append(DeleteResponse(id=id))
                    return len(rs) > 1 and rs or rs[0]

            methods["delete"] = delete

        return methods


def get_anonymous_user() -> str:
    return "anonymous"


class PublicAPI(API[str, Model, CreateForm, UpdateForm]):
    def __init__(
        self,
        model: Type[Model],
        create_form: Type[CreateForm],
        update_form: Type[UpdateForm],
        enable_create: bool = True,
        enable_update: bool = True,
        enable_delete: bool = True,
        enable_get: bool = True,
    ) -> None:
        super().__init__(
            model,
            create_form=create_form,
            update_form=update_form,
            get_current_user=get_anonymous_user,
            enable_create=enable_create,
            enable_update=enable_update,
            enable_delete=enable_delete,
            enable_get=enable_get,
        )



import pytest
from faker import Faker
from pydantic import ValidationError
from tortoise import exceptions as tortoiseExceptions

from anyforce.api import exceptions
from anyforce.model.enum import EnumMissingError

from .model import CharEnum, Model1


@pytest.mark.asyncio
async def test_exceptions(faker: Faker):
    handlers = exceptions.handlers()

    validationErrorHandle = handlers[0][1]
    await validationErrorHandle(
        None, ValidationError([[RuntimeError(faker.pystr())]], Model1.detail())
    )

    ormException = handlers[1][1]
    await ormException(None, tortoiseExceptions.ValidationError(faker.pystr()))
    await ormException(None, tortoiseExceptions.DoesNotExist(faker.pystr()))
    await ormException(None, tortoiseExceptions.IntegrityError(faker.pystr()))
    try:
        await ormException(None, tortoiseExceptions.DBConnectionError(faker.pystr()))
    except tortoiseExceptions.DBConnectionError:
        pass

    try:
        CharEnum("c")
    except EnumMissingError as e:
        enumMissingErrorHandle = handlers[2][1]
        await enumMissingErrorHandle(None, e)



import logging
import logging.handlers
import os.path

from pythonjsonlogger import jsonlogger

from ..json import fast_dumps as dumps


def get(
    path: str, maxBytes: int = 512 * 1024 * 1024, backupCount: int = 10
) -> logging.Logger:
    name = os.path.basename(path)
    logger = logging.getLogger(name)
    logHandler = logging.handlers.RotatingFileHandler(
        path, maxBytes=maxBytes, backupCount=backupCount
    )
    formatter = jsonlogger.JsonFormatter(
        json_serializer=dumps, reserved_attrs=[], timestamp="timestamp"
    )
    logHandler.setFormatter(formatter)
    logger.addHandler(logHandler)
    return logger



from typing import Any, Optional

import pytest
from fastapi import APIRouter, FastAPI
from pydantic import AnyUrl, EmailStr

from anyforce.api import PublicAPI

from .model import Model2, name

pytest_plugins = [
    "anyforce.test.fixtures",
]


@pytest.fixture(scope="session")
def models():
    return [name]


@pytest.fixture(scope="session")
def router(app: FastAPI):
    class CreateForm(Model2.form()):
        text_field: AnyUrl

    class UpdateForm(Model2.detail(required_override=False)):
        text_field: Optional[EmailStr]

    class API(PublicAPI[Model2, CreateForm, UpdateForm]):
        def __init__(self) -> None:
            super().__init__(Model2, CreateForm, UpdateForm)

        async def before_create(
            self, user: str, obj: Model2, input: CreateForm
        ) -> Model2:
            obj = await super().before_create(user, obj, input)
            raw = input.dict(exclude_unset=True)
            foreign_field = raw.pop("foreign_field", None)
            if not foreign_field:
                raw.pop("m2m_field", None)
                obj = Model2(**raw)
            return obj

        async def after_create(self, user: str, obj: Model2, input: CreateForm) -> Any:
            obj = await super().after_create(user, obj, input)
            if obj.id == 1:
                return Model2.detail().from_orm(obj)

        async def before_update(
            self, user: str, obj: Model2, input: UpdateForm
        ) -> Model2:
            obj = await super().before_update(user, obj, input)
            raw = input.dict(exclude_unset=True)
            foreign_field = raw.pop("foreign_field", None)
            if not foreign_field:
                new_obj = await Model2.filter(id=obj.id).first()
                assert new_obj
                raw.pop("m2m_field", None)
                new_obj.update(raw)
                obj = new_obj
            return obj

        async def after_update(self, user: str, obj: Model2, input: UpdateForm) -> Any:
            obj = await super().after_update(user, obj, input)
            if obj.id == 1:
                return Model2.detail().from_orm(obj)

        async def before_delete(self, user: str, obj: Model2) -> Any:
            obj = await super().before_delete(user, obj)
            new_obj = await Model2.filter(id=obj.id).first()
            assert new_obj
            return new_obj

    router = APIRouter(prefix="/models")
    API().bind(router)
    app.include_router(router)
    return router



from datetime import datetime
from typing import Any, Dict, Optional

from tortoise import fields
from tortoise.backends.base.client import BaseDBAsyncClient

from .base import BaseModel


class RecoverableModel(BaseModel):

    is_deleted = fields.BooleanField(null=False, description="是否已经删除", default=False)
    delete_or_recover_at = fields.DatetimeField(null=True, description="删除时间")

    class meta:
        abstract = True

    def update(self, input: Any):
        dic: Dict[str, Any] = (
            input if isinstance(input, dict) else input.dict(exclude_unset=True)
        )
        is_deleted = dic.get("is_deleted")
        if is_deleted is not None and self.is_deleted != is_deleted:
            dic["delete_or_recover_at"] = datetime.now()
        return super().update(dic)

    async def delete(self, using_db: Optional[BaseDBAsyncClient] = None) -> None:
        self.update(
            {
                "is_deleted": True,
                "delete_or_recover_at": datetime.now(),
            }
        )
        await self.save()
        return



import re
from typing import Any, List, Match, Optional, Tuple, Union

import inflection
from fastapi import FastAPI, HTTPException, Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.responses import ORJSONResponse
from pydantic import (
    AnyStrMaxLengthError,
    EmailError,
    IntegerError,
    ListError,
    MissingError,
    UrlError,
    ValidationError,
)
from pydantic.error_wrappers import ErrorWrapper
from tortoise import exceptions

from ..logging import getLogger
from ..model import BaseModel
from ..model.enum import EnumMissingError

logger = getLogger(__name__)

HTTPForbiddenError = HTTPException(
    status_code=status.HTTP_403_FORBIDDEN, detail={"errors": "禁止访问"}
)
HTTPNotFoundError = HTTPException(
    status_code=status.HTTP_404_NOT_FOUND, detail={"errors": "不存在"}
)
HTTPUnAuthorizedError = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED, detail={"errors": "未认证"}
)
HTTPPreconditionRequiredError = HTTPException(
    status_code=status.HTTP_428_PRECONDITION_REQUIRED, detail={"errors": "请求数据已过期"}
)


ValidateError = Union[RequestValidationError, ErrorWrapper, ValidationError]


def translate_validation_error(e: ValidationError) -> List[str]:
    msgs: List[str] = []
    model: Any = getattr(e, "model")
    raw_errors: List[Any] = getattr(e, "raw_errors")
    for raw_error in raw_errors:
        log = logger.with_field(raw_error=raw_error, raw_error_type=type(raw_error))
        if isinstance(raw_error, ErrorWrapper):
            inner_exc = getattr(raw_error, "exc")

            if isinstance(inner_exc, ValidationError):
                msgs = msgs + translate_validation_error(inner_exc)
                continue

            translated_msg = ""
            if isinstance(inner_exc, MissingError):
                translated_msg = "是必填项"
            elif isinstance(inner_exc, IntegerError):
                translated_msg = "为无效的整数"
            elif isinstance(inner_exc, EmailError):
                translated_msg = "为无效的邮箱地址"
            elif isinstance(inner_exc, UrlError):
                translated_msg = "为无效的链接"
            elif isinstance(inner_exc, AnyStrMaxLengthError):
                translated_msg = "数据过长"
            elif isinstance(inner_exc, ListError):
                continue
            else:
                # TODO: tranlstate more if needed
                log.with_field(
                    inner_exc=inner_exc, inner_exc_type=type(inner_exc)
                ).warn("not translate")

            if translated_msg:
                for loc in raw_error.loc_tuple():
                    property = model.schema().get("properties", {}).get(loc, {})
                    title: str = (
                        property.get("description") or property.get("title") or str(loc)
                    )
                    msgs.append(f"{title} {translated_msg}")
                continue
        else:
            log.warn("not translate")
        msgs.append(str(raw_error))
    return msgs


def translate_validate_error(
    e: Union[ValidationError, List[ValidationError]]
) -> List[str]:
    msgs: List[str] = []
    errors: List[Any] = e if isinstance(e, List) else [e]
    for error in errors:
        msgs += translate_validation_error(error)
    return msgs


def translate_orm_error(
    e: exceptions.BaseORMException,
) -> Tuple[int, List[str]]:
    if isinstance(e, exceptions.IntegrityError):
        status_code = status.HTTP_400_BAD_REQUEST
        args: List[str] = []
        for arg in e.args:
            regxps = [
                "UNIQUE constraint failed: (\\S+)\\.(\\S+)",
                ".+Duplicate entry '(\\S+)' for key '(\\S+)\\.(\\S+)'.+",
            ]
            matched: Optional[Match[str]] = None
            for regxp in regxps:
                matched = re.match(regxp, str(arg))
                if matched:
                    groups = matched.groups()
                    if len(groups) == 2:
                        groups = ("", *groups)
                    value = groups[0]
                    model_name = inflection.camelize(groups[1])
                    field_name = groups[2]
                    model, field = BaseModel.get_field(model_name, field_name)
                    field_description = (
                        field
                        and getattr(field, "description", None)
                        or inflection.humanize(field_name)
                    )
                    meta = model and getattr(model, "Meta", None)
                    model_title = (
                        meta
                        and getattr(
                            meta, "title", model and inflection.humanize(model.__name__)
                        )
                        or "对象"
                    )
                    status_code = status.HTTP_409_CONFLICT
                    arg = f"{field_description} 为 {value} 的{model_title}已存在"
                    break
            if not matched:
                logger.with_field(arg=arg, type=type(arg)).warn("not translate")
            args.append(str(arg))
        return status_code, args
    if isinstance(e, exceptions.ValidationError):
        return status.HTTP_400_BAD_REQUEST, [str(args) for args in e.args]
    if isinstance(e, exceptions.DoesNotExist):
        return status.HTTP_404_NOT_FOUND, ["不存在"]
    raise e


def handlers():
    async def validationErrorHandle(
        request: Optional[Request], exc: ValidationError
    ) -> ORJSONResponse:
        return ORJSONResponse(
            {"detail": {"errors": translate_validate_error(exc)}},
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    async def ormException(
        request: Optional[Request], exc: exceptions.BaseORMException
    ) -> ORJSONResponse:
        status_code, msgs = translate_orm_error(exc)
        return ORJSONResponse(
            {"detail": {"errors": msgs}},
            status_code=status_code,
        )

    async def enumMissingErrorHandle(
        request: Optional[Request], exc: EnumMissingError
    ) -> ORJSONResponse:
        return ORJSONResponse(
            {"detail": {"errors": [str(args) for args in exc.args]}},
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    return (
        ([RequestValidationError, ValidationError], validationErrorHandle),
        ([exceptions.BaseORMException], ormException),
        ([EnumMissingError], enumMissingErrorHandle),
    )


def register(app: FastAPI):
    for (errors, handler) in handlers():
        for e in errors:
            app.exception_handler(e)(handler)  # type: ignore



from typing import Any

from starlette_context import context


def get(k: str) -> Any:
    return context.data.get(k)  # type: ignore


def set(k: str, v: Any):
    context.data[k] = v  # type: ignore


def user():
    return get("user")


def set_user(user: Any):
    return set("user", user)



import logging
import os
from typing import Optional, cast

import colorama
from pythonjsonlogger import jsonlogger

from . import colorful
from .context import ContextLogger
from .level import SUCCESS

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL


def patch() -> None:
    if logging.getLoggerClass() == ContextLogger:
        return

    os.environ["PYCHARM_HOSTED"] = "true"
    colorama.init()

    logging.addLevelName(SUCCESS, "SUCCESS")
    logging.setLoggerClass(ContextLogger)

    logger = logging.getLogger()
    logger.setLevel(logging.getLevelName(os.environ.get("LOGLEVEL", "INFO").upper()))
    logHandler = logging.StreamHandler()
    formatter = jsonlogger.JsonFormatter(
        "%(levelname)s %(filename) %(lineno)s %(message)s",
        json_serializer=colorful.dumps,
    )
    logHandler.setFormatter(formatter)
    logger.addHandler(logHandler)


def getLogger(name: Optional[str] = None) -> ContextLogger:
    patch()
    return cast(ContextLogger, logging.getLogger(name))



from __future__ import annotations

from typing import Any

from aenum import IntEnum as aIntEnum
from aenum import StrEnum as aStrEnum


class EnumMissingError(Exception):
    def __init__(self, enum_type: Any, value: Any, *args: Any) -> None:
        super().__init__(*args)
        self.enum_type = enum_type
        self.value = value


class StrEnum(aStrEnum):
    _init_ = "value __doc__"

    @classmethod
    def _missing_(cls, value: Any):
        raise EnumMissingError(cls, value, f"枚举值 {value.__doc__ or value} 不存在")


class IntEnum(aIntEnum):
    _init_ = "value __doc__"

    @classmethod
    def _missing_(cls, value: Any):
        raise EnumMissingError(cls, value, f"枚举值 {value.__doc__} 不存在")



from typing import Any, Dict, List, Optional, Union

import requests
from requests.utils import dict_from_cookiejar

from ..json import raw_dumps


class Response(object):
    def __init__(self, r: requests.Response) -> None:
        self.r = r

    @property
    def content(self) -> bytes:
        return self.r.content

    @property
    def cookies(self) -> Dict[str, str]:
        return dict_from_cookiejar(self.r.cookies)  # type: ignore

    @property
    def text(self) -> str:
        return self.r.text

    @property
    def raw(self) -> Any:
        return self.r.raw

    @property
    def status_code(self) -> int:
        return self.r.status_code

    def json(self, **kwargs: Any) -> Union[Dict[str, Any], List[Any]]:
        return self.r.json(**kwargs)  # type: ignore

    def json_array(self, **kwargs: Any) -> List[Any]:
        r = self.json(**kwargs)
        assert not isinstance(r, dict)
        return r

    def json_object(self, **kwargs: Any) -> Dict[str, Any]:
        r = self.json(**kwargs)
        assert isinstance(r, dict)
        return r


def request(
    method: str,
    url: str,
    params: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, Any]] = None,
    json: Optional[Any] = None,
    session: Optional[requests.Session] = None,
    allow_redirects: bool = True,
    *args: Any,
    **kwargs: Any,
) -> Response:
    request_session = session or requests.session()
    try:
        data = kwargs.pop("data", None)
        headers = headers or {}
        if json and not data:
            data = raw_dumps(json)
            headers["Content-Type"] = "application/json"
        r = request_session.request(  # type: ignore
            method,
            url,
            headers=headers,
            params=params,
            data=data,
            allow_redirects=allow_redirects,
            *args,
            **kwargs,
        )
        return Response(r)
    finally:
        if session is None:
            request_session.close()


def get(
    url: str,
    params: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, Any]] = None,
    session: Optional[requests.Session] = None,
    *args: Any,
    **kwargs: Any,
) -> Response:
    return request(
        "GET",
        url,
        params=params,
        headers=headers,
        session=session,
        *args,
        **kwargs,
    )


def post(
    url: str,
    params: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, Any]] = None,
    json: Optional[Any] = None,
    session: Optional[requests.Session] = None,
    *args: Any,
    **kwargs: Any,
) -> Response:
    return request(
        "POST",
        url,
        params=params,
        headers=headers,
        json=json,
        session=session,
        *args,
        **kwargs,
    )


def put(
    url: str,
    params: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, Any]] = None,
    json: Optional[Any] = None,
    session: Optional[requests.Session] = None,
    *args: Any,
    **kwargs: Any,
) -> Response:
    return request(
        "PUT",
        url,
        params=params,
        headers=headers,
        json=json,
        session=session,
        *args,
        **kwargs,
    )


def delete(
    url: str,
    params: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, Any]] = None,
    session: Optional[requests.Session] = None,
    *args: Any,
    **kwargs: Any,
) -> Response:
    return request(
        "DELETE",
        url,
        params=params,
        headers=headers,
        session=session,
        *args,
        **kwargs,
    )



from datetime import datetime, timedelta

from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer
from jose import jwt as jwt_lib

from ..exceptions import HTTPUnAuthorizedError


def gen(
    token_url: str,
    secret: str,
    expire_after_seconds: int = 3600 * 24 * 30,
    algorithm: str = "HS256",
):
    oauth2_scheme = OAuth2PasswordBearer(tokenUrl=token_url)

    async def get_current_user(
        token: str = Depends(oauth2_scheme),
    ) -> str:
        payload = jwt_lib.decode(token, secret, algorithms=[algorithm])
        user_id: str = payload.get("sub", "")
        if not user_id:
            raise HTTPUnAuthorizedError
        exp = float(payload.get("exp", 0))
        assert exp > datetime.now().timestamp()
        return user_id

    def authorize(user_id: str) -> str:
        exp = datetime.now() + timedelta(seconds=expire_after_seconds)
        token = jwt_lib.encode(
            {"sub": user_id, "exp": int(exp.timestamp())},
            secret,
            algorithm=algorithm,
        )
        return token

    return get_current_user, authorize



from typing import Any, Dict, List

from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.sessions import SessionMiddleware
from starlette_context.middleware import RawContextMiddleware
from tortoise.contrib.fastapi import register_tortoise

from .api.exceptions import register


def create_app(
    secret_key: str,
    https_only: bool,
    allow_origins: List[str],
    tortoise_config: Dict[str, Any],
):
    app = FastAPI()
    app.add_middleware(RawContextMiddleware)
    app.add_middleware(SessionMiddleware, secret_key=secret_key, https_only=https_only)
    app.add_middleware(
        CORSMiddleware,
        allow_origins=allow_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    register(app)
    register_tortoise(
        app,
        config=tortoise_config,
        generate_schemas=False,
        add_exception_handlers=False,
    )

    @app.get("/healthz")
    async def _() -> str:
        return ""

    return app



from typing import Any, Dict

from fastapi import status

from anyforce.json import fast_dumps
from anyforce.test import TestAPI as Base


class TestAPI(Base):
    def endpoint(self) -> str:
        return "/models"

    def create_tests(self):
        unique_field = self.faker.pystr()
        yield {
            "char_enum_field": "b",
            "required_char_field": self.faker.pystr(),
            "nullable_char_field": unique_field,
            "int_field": self.faker.pyint(),
            "bigint_field": self.faker.pyint(),
            "text_field": self.faker.url(),
            "json_field": list(self.faker.pytuple()),
        }, status.HTTP_201_CREATED, None

        # int
        yield {
            "char_enum_field": "b",
            "required_char_field": self.faker.pystr(),
            "int_field": self.faker.name(),
            "bigint_field": self.faker.pyint(),
            "text_field": self.faker.url(),
            "json_field": list(self.faker.pytuple()),
        }, status.HTTP_400_BAD_REQUEST, None

        # url
        yield {
            "char_enum_field": "b",
            "required_char_field": self.faker.pystr(),
            "int_field": self.faker.name(),
            "bigint_field": self.faker.pyint(),
            "text_field": self.faker.name(),
            "json_field": list(self.faker.pytuple()),
        }, status.HTTP_400_BAD_REQUEST, None

        # enum
        yield {
            "char_enum_field": "c",
            "required_char_field": self.faker.pystr(),
            "int_field": self.faker.pyint(),
            "bigint_field": self.faker.pyint(),
            "text_field": self.faker.url(),
            "json_field": list(self.faker.pytuple()),
        }, status.HTTP_400_BAD_REQUEST, None

        # unique
        yield {
            "char_enum_field": "a",
            "required_char_field": self.faker.pystr(),
            "nullable_char_field": unique_field,
            "int_field": self.faker.pyint(),
            "bigint_field": self.faker.pyint(),
            "text_field": self.faker.url(),
            "json_field": list(self.faker.pytuple()),
        }, status.HTTP_409_CONFLICT, None

        # required
        yield {
            "char_enum_field": "a",
            "nullable_char_field": self.faker.pystr(),
            "int_field": self.faker.pyint(),
            "bigint_field": self.faker.pyint(),
            "text_field": self.faker.url(),
            "json_field": list(self.faker.pytuple()),
        }, status.HTTP_400_BAD_REQUEST, None

        # too lang
        yield {
            "char_enum_field": "a",
            "required_char_field": self.faker.pystr(min_chars=33, max_chars=64),
            "nullable_char_field": self.faker.pystr(),
            "int_field": self.faker.pyint(),
            "bigint_field": self.faker.pyint(),
            "text_field": self.faker.url(),
            "json_field": list(self.faker.pytuple()),
        }, status.HTTP_400_BAD_REQUEST, None

        # save related
        yield {
            "char_enum_field": "a",
            "required_char_field": self.faker.pystr(),
            "nullable_char_field": self.faker.pystr(),
            "int_field": self.faker.pyint(),
            "bigint_field": self.faker.pyint(),
            "text_field": self.faker.url(),
            "json_field": list(self.faker.pytuple()),
            "foreign_field": {"name": "name"},
        }, status.HTTP_201_CREATED, None
        yield {
            "char_enum_field": "a",
            "required_char_field": self.faker.pystr(),
            "nullable_char_field": self.faker.pystr(),
            "int_field": self.faker.pyint(),
            "bigint_field": self.faker.pyint(),
            "text_field": self.faker.url(),
            "json_field": list(self.faker.pytuple()),
            "foreign_field": {"name": "name"},
            "m2m_field": [
                {"name": self.faker.pystr()},
                {"name": self.faker.pystr()},
            ],
        }, status.HTTP_201_CREATED, None

    def list_tests(self):
        def check_computed(params: Dict[str, Any], r: Any) -> None:
            for e in r["data"]:
                assert (
                    e["int_field_plus_bigint_field"]
                    == e["int_field"] + e["bigint_field"]
                )
                assert (
                    e["async_int_field_plus_bigint_field"]
                    == e["int_field"] + e["bigint_field"]
                )

        # prefetch / computed / order_by
        yield {
            "prefetch": [
                "foreign_field",
                "m2m_field",
                "int_field_plus_bigint_field",
                "async_int_field_plus_bigint_field",
            ],
            "order_by": "int_field",
        }, status.HTTP_200_OK, check_computed

        def check1(params: Dict[str, Any], r: Any) -> None:
            assert r["total"] == 1
            assert r["data"][0]["char_enum_field"] == "b"

        # filter
        yield {
            "condition": [fast_dumps({"char_enum_field": "b"})]
        }, status.HTTP_200_OK, check1

        yield {
            "condition": [fast_dumps({"char_enum_field.in": ["b"]})]
        }, status.HTTP_200_OK, check1

        def check2(params: Dict[str, Any], r: Any) -> None:
            assert r["total"] == 2

        yield {
            "condition": [fast_dumps({"foreign_field.name": "name"})]
        }, status.HTTP_200_OK, check2

    def get_tests(self):
        yield {
            "id": 1,
            "prefetch": ["int_field_plus_bigint_field"],
        }, status.HTTP_200_OK, None
        yield {"id": 10000}, status.HTTP_404_NOT_FOUND, None

    def update_tests(self):
        yield {
            "id": 1,
            "prefetch": ["int_field_plus_bigint_field"],
            "body": {
                "text_field": self.faker.email(),
                "bigint_field": self.faker.pyint(),
                "m2m_field": [{"id": 1, "name": self.faker.pystr()}],
            },
        }, status.HTTP_200_OK, None

        yield {
            "id": 1,
            "prefetch": ["int_field_plus_bigint_field"],
            "body": {"text_field": self.faker.name()},
        }, status.HTTP_400_BAD_REQUEST, None

        yield {
            "id": 2,
            "prefetch": ["int_field_plus_bigint_field"],
            "body": {"bigint_field": self.faker.pyint()},
        }, status.HTTP_200_OK, None

        yield {
            "id": 10000,
            "body": {
                "bigint_field": self.faker.pyint(),
            },
        }, status.HTTP_404_NOT_FOUND, None

    def delete_tests(self):
        yield {"id": 1}, status.HTTP_200_OK, None



from datetime import datetime

try:
    from ciso8601 import parse_datetime  # type: ignore
except ImportError:
    parse_datetime = datetime.fromisoformat
from typing import Any, Dict, List, Union, cast

import orjson
from fastapi.encoders import jsonable_encoder


def fast_dumps(o: Any) -> str:
    return orjson.dumps(o).decode("utf-8")


def raw_dumps(o: Any) -> bytes:
    return orjson.dumps(o, default=jsonable_encoder)


def dumps(o: Any) -> str:
    return orjson.dumps(o, default=jsonable_encoder).decode("utf-8")


def parse_iso_datetime(s: str) -> datetime:
    return cast(datetime, parse_datetime(s))


def decoder(input: Any) -> Any:
    if isinstance(input, dict):
        input = cast(Dict[Any, Any], input)
        for k, v in input.items():
            input[k] = decoder(v)
    elif isinstance(input, List):
        input = cast(List[Any], input)
        for i, v in enumerate(input):
            input[i] = decoder(v)
    elif isinstance(input, str) and input.find("T") == 10:
        try:
            return parse_iso_datetime(input)
        except ValueError:
            pass
    return input


def loads(raw: Union[bytes, bytearray, str]):
    o = orjson.loads(raw)
    return decoder(o)



import inspect
from datetime import datetime
from functools import lru_cache
from typing import Any, Dict, List, Optional, Set, Tuple, Type, cast

from tortoise import Tortoise, fields
from tortoise.backends.base.client import BaseDBAsyncClient
from tortoise.contrib.pydantic import PydanticModel, pydantic_model_creator
from tortoise.contrib.pydantic.creator import PydanticMeta
from tortoise.models import Field, MetaInfo, Model

from .patch import patch_pydantic


class BaseModel(Model):
    id: int = fields.IntField(pk=True)
    created_at: datetime = fields.DatetimeField(null=False, auto_now_add=True)
    updated_at: datetime = fields.DatetimeField(null=False, auto_now=True)

    class Meta:
        abstract = True

    # class PydanticMeta:
    #     max_levels = 1  # 最大模型层级
    #     computed: Tuple[str, ...] = ()  # 计算量, 异步计算量返回值需要标记为 Optional
    #     list_exclude: Tuple[str, ...] = ()  # 列表排除
    #     detail_include: Tuple[str, ...] = ()  # 详情叠加
    #     form_exclude: Tuple[str, ...] = ()  # 表单排除

    class FormPydanticMeta:
        computed = []

    async def dict(self, prefetch: Optional[List[str]] = None) -> Dict[str, Any]:
        if prefetch:
            await self.fetch_related(*prefetch)
        return self.detail().from_orm(self).dict()

    @classmethod
    @lru_cache
    def list(cls) -> Type[PydanticModel]:
        meta: Optional[PydanticMeta] = getattr(cls, "PydanticMeta", None)
        list_exclude: Tuple[str, ...] = meta and getattr(meta, "list_exclude", ()) or ()
        return cls.make_pydantic(
            name="list", exclude=list_exclude, required_override=False
        )

    @classmethod
    @lru_cache
    def detail(
        cls,
        required_override: Optional[bool] = None,
        from_models: Tuple[str, ...] = (),
    ) -> Type[PydanticModel]:
        meta: Optional[PydanticMeta] = getattr(cls, "PydanticMeta", None)
        detail_include: Tuple[str, ...] = (
            meta and getattr(meta, "detail_include", ()) or ()
        )
        return cls.make_pydantic(
            name="detail",
            include=detail_include,
            required_override=required_override,
            from_models=from_models,
        )

    @classmethod
    @lru_cache
    def form(
        cls,
        required_override: Optional[bool] = None,
        from_models: Tuple[str, ...] = (),
    ) -> Type[PydanticModel]:
        meta: Optional[PydanticMeta] = getattr(cls, "PydanticMeta", None)
        form_exclude: Tuple[str, ...] = meta and getattr(meta, "form_exclude", ()) or ()
        return cls.make_pydantic(
            name="form",
            exclude=(
                *form_exclude,
                *(
                    ["id", "created_at", "updated_at"]
                    if required_override is None
                    else ["id", "created_at"]
                ),
            ),
            required_override=required_override,
            from_models=from_models,
            is_form=True,
        )

    @classmethod
    def make_pydantic(
        cls,
        name: str,
        include: Optional[Tuple[str, ...]] = None,
        exclude: Optional[Tuple[str, ...]] = None,
        required_override: Optional[bool] = None,
        from_models: Tuple[str, ...] = (),
        is_form: bool = False,
    ):
        parts = [cls.__module__, cls.__qualname__, name]
        if from_models:
            parts.append("in")
            parts += from_models
        if required_override is not None:
            parts.append("required" if required_override else "optional")

        meta: Optional[PydanticMeta] = getattr(cls, "PydanticMeta", None)
        max_levels: int = meta and getattr(meta, "max_levels", None) or 1
        return patch_pydantic(
            pydantic_model_creator(
                cls,
                name=".".join(parts),
                include=include or (),
                exclude=exclude or (),
                meta_override=cls.FormPydanticMeta if is_form else None,
            ),
            from_models=(*from_models, cls.__qualname__),
            required_override=required_override,
            is_form=is_form,
            max_levels=max_levels,
        )

    def update(self, input: Any):
        dic: Dict[str, Any] = (
            input if isinstance(input, dict) else input.dict(exclude_unset=True)
        )
        self.update_from_dict(dic)  # type: ignore

    async def fetch_related(
        self, *args: Any, using_db: Optional[BaseDBAsyncClient] = None
    ) -> None:
        meta: Optional[PydanticMeta] = getattr(self.__class__, "PydanticMeta", None)
        computed: Set[str]
        if meta and hasattr(meta, "computed"):
            computed = set(meta.computed)
            for field in args:
                if field in computed:
                    f = getattr(self, field, None)
                    if not f:
                        continue
                    if inspect.iscoroutinefunction(f):
                        setattr(self, field, await f())
        else:
            computed = set()

        normlized_args = [
            self.normalize_field(field) if isinstance(field, str) else field
            for field in args
            if field not in computed
        ]
        return await super().fetch_related(*normlized_args, using_db=using_db)

    @classmethod
    async def save_related(cls, raw: Any) -> Tuple[Any, Dict[str, Any]]:
        meta: MetaInfo = getattr(cls, "_meta")

        # 保存 ForeignKeyField
        fk_fields: Set[str] = meta.fk_fields
        for fk_field_name in fk_fields:
            v: Optional[Any] = raw.pop(fk_field_name, None)
            if not v:
                continue

            field_model: Type[BaseModel] = cast(
                Type[BaseModel], cls.get_field_model(fk_field_name)
            )

            v_id = v.pop("id", None)
            if v_id:
                field_value: Optional[BaseModel] = await field_model.filter(
                    id=v_id
                ).first()
                assert field_value
                field_value.update(v)
            else:
                field_value = field_model(**v)
            await field_value.save()
            raw[fk_field_name] = field_value

        # 保存 ManyToMany
        m2m_raw: Dict[str, Any] = {}
        m2m_fields: Set[str] = meta.m2m_fields
        for m2m_field_name in m2m_fields:
            v: Optional[Any] = raw.pop(m2m_field_name, None)
            if not v:
                continue
            assert isinstance(v, List)
            vs: List[Any] = v

            field_model: Type[BaseModel] = cast(
                Type[BaseModel], cls.get_field_model(m2m_field_name)
            )

            field_values: List[BaseModel] = []
            for v in vs:
                assert v
                v_id = v.pop("id", None)
                if v_id:
                    field_value: Optional[BaseModel] = await field_model.filter(
                        id=v_id
                    ).first()
                    assert field_value
                    field_value.update(v)
                else:
                    field_value = field_model(**v)
                await field_value.save()
                field_values.append(field_value)

            if field_values:
                m2m_raw[m2m_field_name] = field_values

        return raw, m2m_raw

    @staticmethod
    def normalize_field(field: str) -> str:
        return field.replace(".", "__")

    @classmethod
    @lru_cache
    def get_model(cls, model_name: str) -> Optional[Type[Model]]:
        parts: List[str] = model_name.split(".")
        if len(parts) == 1:
            for app in Tortoise.apps:
                m = cls.get_model(f"{app}.{model_name}")
                if m:
                    return m
        return cast(
            Optional[Type[BaseModel]],
            Tortoise.apps.get(parts[0], {}).get(parts[1], None),
        )

    @classmethod
    @lru_cache
    def get_field(
        cls, model_name: str, field_name: str
    ) -> Tuple[Optional[Type[Model]], Optional[Field]]:
        model = cls.get_model(model_name)
        meta: Optional[MetaInfo] = model and getattr(model, "_meta")
        field = model and meta and meta.fields_map.get(field_name)
        return model, field

    @classmethod
    @lru_cache
    def get_field_model(cls, field: str) -> Type[Model]:
        meta: MetaInfo = getattr(cls, "_meta")
        fields_map = meta.fields_map
        fk_field = fields_map[field]
        model_name = getattr(fk_field, "model_name")
        model = cls.get_model(model_name)
        assert model
        return model



from typing import Any

from email_validator import EmailNotValidError, validate_email
from tortoise.exceptions import ValidationError


def email(value: Any):
    try:
        validate_email(str(value))
    except EmailNotValidError:
        raise ValidationError(f"{value} 不是有效的邮箱地址")



from typing import Any, Callable, Dict, Optional
from urllib.parse import urlencode, urljoin

import aiohttp
from fastapi import APIRouter, Request
from starlette.responses import RedirectResponse

from ... import json
from ..exceptions import HTTPUnAuthorizedError


class OAuth2(object):
    def __init__(
        self,
        base_url: str,
        client_id: str,
        client_secret: str,
        scheme: Optional[str] = None,
    ):
        super().__init__()
        self.base_url = base_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.scheme = scheme

    def join(self, path: str, q: Dict[str, str] = {}):
        return f"{urljoin(self.base_url, path)}?{urlencode(q)}"

    def redirect_url(self, request: Request, redirect_uri: str):
        base = str(
            request.url.replace(scheme=self.scheme) if self.scheme else request.url
        )
        url = f"{urljoin(base, 'auth')}?{urlencode({'redirect_uri': redirect_uri})}"
        return url

    def auth_url(self, redirect_uri: str):
        return self.join(
            "auth",
            {
                "client_id": self.client_id,
                "response_type": "code",
                "redirect_uri": redirect_uri,
            },
        )

    async def auth(self, code: str, redirect_uri: str):
        async with aiohttp.ClientSession() as session:
            async with session.post(
                self.join("token"),
                data=aiohttp.FormData(
                    {
                        "client_id": "dashboard",
                        "client_secret": "f1ace9da-0d6a-4532-938d-8c8ea7730487",
                        "grant_type": "authorization_code",
                        "code": code,
                        "redirect_uri": redirect_uri,
                    }
                ),
            ) as response:
                r = await response.json(loads=json.loads)
                return r["access_token"]

    async def userinfo(self, token: str):
        async with aiohttp.ClientSession() as session:
            async with session.get(
                self.join("userinfo"), headers={"Authorization": f"Bearer {token}"}
            ) as response:
                return await response.json(loads=json.loads)

    def bind(
        self, router: APIRouter, wrap: bool, verify: Callable[[Request, Any, str], Any]
    ):
        @router.get("/login")
        def login(request: Request, redirect_uri: str = ""):
            redirect_uri = (
                self.redirect_url(request, redirect_uri)
                if not redirect_uri or wrap
                else redirect_uri
            )
            return RedirectResponse(self.auth_url(redirect_uri))

        @router.get("/auth")
        async def auth(request: Request, code: str, redirect_uri: str = ""):
            token = await self.auth(
                code,
                self.redirect_url(request, redirect_uri)
                if not redirect_uri or wrap
                else redirect_uri,
            )
            r = await verify(request, await self.userinfo(token), redirect_uri)
            return r if r else HTTPUnAuthorizedError

        return login, auth



from typing import Optional

from tortoise import fields
from tortoise.fields.relational import ManyToManyRelation

from anyforce.model import BaseModel, StrEnum
from anyforce.model import fields as afields


class CharEnum(StrEnum):
    a = "a", "a"
    b = "b", "b"


class User(BaseModel):
    email = fields.CharField(64, index=True, null=False, description="邮箱", default="")
    hashed_password = fields.CharField(128, null=False, default="")


class Model1(BaseModel):
    name = fields.CharField(max_length=32)


class Model2(BaseModel):
    title = "测试"

    int_field = fields.IntField(title="数字字段")
    bigint_field = fields.BigIntField()
    char_enum_field = fields.CharEnumField(CharEnum)
    nullable_char_field = fields.CharField(
        max_length=32, unique=True, null=True, description="唯一字段"
    )
    default_char_field = fields.CharField(max_length=32, null=False, default="")
    required_char_field = fields.CharField(
        max_length=32, null=False, description="必填字段"
    )
    text_field = fields.TextField(default="")
    uuid_field = fields.UUIDField(null=True)
    float_field = fields.FloatField(default=0)
    date_field = fields.DateField(null=True)
    datetime_field = fields.DatetimeField(null=True)
    auto_now_field = fields.DatetimeField(null=True, auto_now=True)
    auto_now_add_field = fields.DatetimeField(null=True, auto_now_add=True)
    timedelta_field = fields.TimeDeltaField(null=True)
    json_field = afields.JSONField(default={})
    binary_field = fields.BinaryField(null=True)
    foreign_field = fields.ForeignKeyField(
        "models.Model1", null=True, related_name="model2"
    )
    m2m_field: ManyToManyRelation[Model1] = afields.ManyToManyField(
        "models.Model1", related_name="model2s"
    )

    class PydanticMeta:
        arbitrary_types_allowed = True
        computed = [
            "int_field_plus_bigint_field",
            "async_int_field_plus_bigint_field",
        ]

    def int_field_plus_bigint_field(self) -> int:
        return self.int_field + self.bigint_field

    async def async_int_field_plus_bigint_field(self) -> Optional[int]:
        return self.int_field + self.bigint_field


name = __name__

__all__ = ["name"]



import logging
from functools import cached_property
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, cast

from faker import Faker
from fastapi import APIRouter
from fastapi.encoders import jsonable_encoder
from fastapi.testclient import TestClient

from .request import delete, get, post, put

logger = logging.getLogger()


TestConfigs = Iterable[
    Tuple[
        Dict[str, Any],
        int,
        Optional[Callable[[Dict[str, Any], Any], None]],
    ]
]


class TestAPI:
    def endpoint(self) -> str:
        raise NotImplementedError()

    @cached_property
    def faker(self) -> Faker:
        return Faker()

    def create_tests(
        self,
    ) -> TestConfigs:
        return []

    def test_create(self, client: TestClient, database: bool, router: APIRouter):
        assert router
        assert database
        for json, status_code, callback in self.create_tests():
            r = post(self.endpoint(), json=json, session=client)
            self.log_request(json, status_code, r)

            obj = r.json_object()
            if status_code < 300:
                self.assert_obj(obj)
                self.log_compare(obj, jsonable_encoder(json))

            callback and callback(json, obj)

    def list_tests(
        self,
    ) -> TestConfigs:
        return []

    def test_list(self, client: TestClient, database: bool, router: APIRouter):
        assert router
        assert database
        for params, status_code, callback in self.list_tests():
            r = get(self.endpoint(), params=params, session=client)
            self.log_request(params, status_code, r)

            r = r.json_object()
            if status_code < 300:
                # validate total offset limit
                assert r
                assert r["total"] > 0
                assert len(r["data"]) > 0 and len(r["data"]) < params.get("limit", 20)
                assert params.get("offset", 0) + len(r["data"]) <= r["total"]

                self.assert_obj(r["data"][0])

                # validate prefetch
                prefetch: Any = params.get("prefetch", [])
                prefetch = prefetch if isinstance(prefetch, list) else [prefetch]
                for k in prefetch:
                    exist = False
                    for e in r["data"]:
                        if e.get(k) is not None:
                            exist = True
                            break
                    if not exist:
                        logger.info(f"prefetch: {prefetch} => {r['data']}")
                    assert exist

            callback and callback(params, r)

    def get_tests(
        self,
    ) -> TestConfigs:
        return []

    def test_get(self, client: TestClient, database: bool, router: APIRouter):
        assert router
        assert database
        for params, status_code, callback in self.get_tests():
            input_params = params.copy()
            r = get(
                f"{self.endpoint()}/{params.pop('id')}",
                params=params,
                session=client,
            )
            self.log_request(input_params, status_code, r)

            obj = r.json_object()
            if status_code < 300:
                self.assert_obj(obj)

                # validate preftch
                prefetch: Any = params.get("prefetch", [])
                prefetch = prefetch if isinstance(prefetch, list) else [prefetch]
                for k in prefetch:
                    assert obj.get(k) is not None

            callback and callback(params, obj)

    def update_tests(
        self,
    ) -> TestConfigs:
        return []

    def test_update(self, client: TestClient, database: bool, router: APIRouter):
        assert router
        assert database
        for params, status_code, callback in self.update_tests():
            input_params = params.copy()
            body = params.pop("body", {})
            r = put(
                f"{self.endpoint()}/{params.pop('id')}",
                json=body,
                params=params,
                session=client,
            )
            self.log_request(input_params, status_code, r)

            obj = r.json_object()
            if status_code < 300:
                self.assert_obj(obj)
                self.log_compare(obj, body)

            callback and callback(params, obj)

    def delete_tests(
        self,
    ) -> TestConfigs:
        return []

    def test_delete(self, client: TestClient, database: bool, router: APIRouter):
        assert router
        assert database
        for params, status_code, callback in self.delete_tests():
            input_params = params.copy()
            r = delete(
                f"{self.endpoint()}/{params.pop('id')}",
                params=params,
                session=client,
            )
            self.log_request(input_params, status_code, r)

            obj = r.json_object()
            if status_code < 300:
                assert obj
                assert obj["id"]

            callback and callback(params, obj)

    @staticmethod
    def log_request(input: Any, status_code: int, r: Any):
        if r.status_code != status_code:
            logging.info(
                f"expect: {status_code}, request: {input} -> [{r.status_code}]{r.text}"
            )
            import pdb

            pdb.set_trace()
        assert r.status_code == status_code

    @staticmethod
    def assert_obj(obj: Any):
        assert obj
        assert obj["id"]
        assert obj["created_at"]
        assert obj["updated_at"]

    @staticmethod
    def log_compare(lv: Any, rv: Any):
        diff: Dict[str, Any] = {}
        r = TestAPI.compare(lv, rv, "", diff)
        if not r:
            logger.info(f"not equal: {diff}")
        assert r

    @staticmethod
    def compare(lv: Any, rv: Any, path: str, diff: Dict[str, Any]):
        if isinstance(lv, dict) and isinstance(rv, dict):
            lv = cast(Dict[str, Any], lv)
            rv = cast(Dict[str, Any], rv)
            for k, v in rv.items():
                if not TestAPI.compare(lv.get(k), v, f"{path}.{k}", diff):
                    return False
            return True
        if isinstance(lv, list) and isinstance(rv, list):
            rv = cast(List[Any], rv)
            for i, v in enumerate(rv):
                if not TestAPI.compare(lv[i], v, f"{path}.{i}", diff):
                    return False
            return True
        if lv == rv:
            return True
        diff[path] = (lv, rv)
        return False



import asyncio
from typing import Iterable

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from tortoise import Tortoise

from ..api import exceptions


@pytest.fixture(scope="session")
def event_loop():
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


async def init_tortoise(models: Iterable[str]):
    await Tortoise.init(  # type: ignore
        db_url="sqlite://:memory:", modules={"models": models}
    )


@pytest.fixture(scope="session")
async def app(models: Iterable[str]):
    await init_tortoise(models)
    app = FastAPI()
    exceptions.register(app)
    yield app


@pytest.fixture(scope="session")
def client(app: FastAPI) -> TestClient:
    return TestClient(app)


@pytest.fixture(scope="module")
async def database(models: Iterable[str]):
    await init_tortoise(models)
    await Tortoise.generate_schemas(False)
    yield True
    await Tortoise._drop_databases()  # type: ignore



# -*- coding: utf-8 -*-

import unittest

from anyforce import logging


class TestLogging(unittest.TestCase):
    def test_logging(self):
        logging.getLogger("test").with_field(k="v").debug("debug")
        logging.getLogger("test").with_field(k="v").info("info")
        logging.getLogger("test").with_field(k="v").warning("warning")
        logging.getLogger("test").with_field(k="v").success("success")
        logging.getLogger("test").with_field(k="v").log(logging.INFO, "info")



from datetime import datetime
from typing import Any, Optional, Type, Union

from tortoise import fields
from tortoise.fields import DatetimeField
from tortoise.fields.relational import ManyToManyRelation
from tortoise.models import Model

from .. import json


class LocalDatetimeField(DatetimeField):
    def to_db_value(
        self, value: Optional[datetime], instance: Union[Type[Model], Model]
    ) -> Optional[datetime]:
        value = super().to_db_value(value, instance)
        if (
            value
            and value.tzinfo is not None
            and value.tzinfo.utcoffset(value) is not None
        ):
            value = value.astimezone()
        return value


def JSONField(**kwargs: Any) -> fields.JSONField:
    return fields.JSONField(encoder=json.fast_dumps, decoder=json.loads, **kwargs)


def ManyToManyField(
    model_name: str,
    through: Optional[str] = None,
    forward_key: Optional[str] = None,
    backward_key: str = "",
    related_name: str = "",
    on_delete: str = fields.CASCADE,
    db_constraint: bool = True,
    **kwargs: Any,
) -> ManyToManyRelation[Any]:
    return fields.ManyToManyField(  # type: ignore
        model_name=model_name,
        through=through,
        forward_key=forward_key,
        backward_key=backward_key,
        related_name=related_name,
        on_delete=on_delete,
        db_constraint=db_constraint,
        **kwargs,
    )



import inspect
from typing import Any, Dict, List, Optional, Tuple, Type, cast

from pydantic import validate_model
from pydantic.fields import SHAPE_LIST, SHAPE_SINGLETON, ModelField
from tortoise import fields
from tortoise.contrib.pydantic import PydanticModel
from tortoise.exceptions import NoValuesFetched
from tortoise.fields.relational import NoneAwaitable
from tortoise.queryset import AwaitableQuery


def patch_pydantic(
    model: Type[PydanticModel],
    from_models: Tuple[str, ...] = (),
    required_override: Optional[bool] = None,
    is_form: bool = False,
    max_levels: int = 1,
) -> Type[PydanticModel]:
    # 解决数据动态加载的问题
    added_fields: Dict[str, ModelField] = {}
    for k, field in model.__fields__.items():
        config = getattr(field.type_, "__config__", None)
        if config:
            orig_model: Optional[Any] = getattr(config, "orig_model", None)
            if orig_model:
                if (
                    orig_model.__qualname__ in from_models
                    or len(from_models) > max_levels
                ):
                    field = ModelField.infer(
                        name=k,
                        value=None,
                        annotation=Optional[Any],
                        class_validators=None,
                        config=config,
                    )
                else:
                    field_pydantic_model: Type[Any] = (
                        orig_model.form(
                            required_override=required_override,
                            from_models=from_models,
                        )
                        if is_form
                        else orig_model.detail(
                            from_models=from_models,
                            required_override=required_override,
                        )
                    )
                    if field.shape == SHAPE_LIST:
                        field = ModelField.infer(
                            name=k,
                            value=None,
                            annotation=Optional[List[field_pydantic_model]],
                            class_validators=None,
                            config=config,
                        )
                    elif field.shape == SHAPE_SINGLETON:
                        field = ModelField.infer(
                            name=k,
                            value=None,
                            annotation=Optional[field_pydantic_model],
                            class_validators=None,
                            config=config,
                        )
                        k_id = f"{k}_id"
                        added_fields[k_id] = ModelField.infer(
                            name=k_id,
                            value=None,
                            annotation=int
                            if (
                                field.required
                                if required_override is None
                                else required_override
                            )
                            else Optional[int],
                            class_validators=None,
                            config=config,
                        )
                model.__fields__[k] = field

        if required_override is not None:
            field.required = required_override
            continue

        if not field.required:
            continue

        field.allow_none = True
        try:
            field.required = not issubclass(field.type_, PydanticModel)
        except TypeError:
            continue

    for k, field in added_fields.items():
        model.__fields__[k] = field

    def from_orm(obj: Any) -> PydanticModel:
        obj = model._decompose_class(obj)  # type: ignore
        new_obj: Dict[str, Any] = {}
        for k, v in obj.items():
            if k not in model.__fields__:
                continue

            if (
                isinstance(v, AwaitableQuery)
                or inspect.iscoroutinefunction(v)
                or v is NoneAwaitable
            ):
                continue
            if isinstance(v, fields.ReverseRelation):
                try:
                    v = cast(Any, v)
                    v: Any = list(v)
                except NoValuesFetched:
                    continue
            new_obj[k] = v

        m = model.__new__(model)
        values, fields_set, validation_error = validate_model(model, new_obj)
        if validation_error:
            raise validation_error
        object.__setattr__(m, "__dict__", values)
        object.__setattr__(m, "__fields_set__", fields_set)
        getattr(m, "_init_private_attributes")()
        return m

    def to_dict(self: model, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        if kwargs and kwargs.get("exclude_none") is not None:
            kwargs["exclude_none"] = True
        return super(model, self).dict(*args, **kwargs)

    model.from_orm = from_orm
    model.dict = to_dict

    return model



from __future__ import annotations

import logging
from typing import Any, Dict, Optional

from .level import SUCCESS


class Context(object):
    def __init__(
        self, logger: logging.Logger, context: Optional[Dict[str, Any]] = None
    ) -> None:
        super(Context, self).__init__()
        self.logger = logger
        self.context: Dict[str, Any] = context or {}

    def get_extras(self, extra: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        context = self.context.copy()
        if extra:
            context.update(extra)

        # 替换特殊关键词
        for keyword in ["name", "level", "msg", "args", "exc_info", "func"]:
            val = context.pop(keyword, None)
            if val:
                context[f"{keyword}_"] = val

        return context

    def with_field(self, **kwargs: Any) -> Context:
        context = self.context.copy()
        context.update(kwargs)
        return Context(self.logger, context=context)

    def log(
        self,
        level: int,
        msg: str,
        *args: Any,
        exc_info: Optional[Exception] = None,
        extra: Optional[Dict[str, Any]] = None,
        stack_info: bool = False,
        stacklevel: int = 1,
    ) -> Context:
        if not self.logger.isEnabledFor(level):
            return self

        self.logger._log(  # type: ignore
            level,
            msg,
            args,
            exc_info=exc_info,
            extra=self.get_extras(extra),
            stack_info=stack_info,
            stacklevel=stacklevel,
        )
        return self

    def debug(
        self,
        msg: str,
        *args: Any,
        exc_info: Optional[Exception] = None,
        extra: Optional[Dict[str, Any]] = None,
        stack_info: bool = False,
    ) -> Context:
        return self.log(
            logging.DEBUG,
            msg,
            exc_info=exc_info,
            extra=extra,
            stack_info=stack_info,
            stacklevel=2,
            *args,
        )

    def info(
        self,
        msg: str,
        *args: Any,
        exc_info: Optional[Exception] = None,
        extra: Optional[Dict[str, Any]] = None,
        stack_info: bool = False,
    ) -> Context:
        return self.log(
            logging.INFO,
            msg,
            exc_info=exc_info,
            extra=extra,
            stack_info=stack_info,
            stacklevel=2,
            *args,
        )

    def warning(
        self,
        msg: str,
        *args: Any,
        exc_info: Optional[Exception] = None,
        extra: Optional[Dict[str, Any]] = None,
        stack_info: bool = False,
    ) -> Context:
        return self.log(
            logging.WARNING,
            msg,
            exc_info=exc_info,
            extra=extra,
            stack_info=stack_info,
            stacklevel=2,
            *args,
        )

    def warn(
        self,
        msg: str,
        *args: Any,
        exc_info: Optional[Exception] = None,
        extra: Optional[Dict[str, Any]] = None,
        stack_info: bool = False,
    ) -> Context:
        return self.log(
            logging.WARNING,
            msg,
            exc_info=exc_info,
            extra=extra,
            stack_info=stack_info,
            stacklevel=2,
            *args,
        )

    def error(
        self,
        msg: str,
        *args: Any,
        exc_info: Optional[Exception] = None,
        extra: Optional[Dict[str, Any]] = None,
        stack_info: bool = False,
    ) -> Context:
        return self.log(
            logging.ERROR,
            msg,
            exc_info=exc_info,
            extra=extra,
            stack_info=stack_info,
            stacklevel=2,
            *args,
        )

    def exception(
        self,
        msg: str,
        *args: Any,
        exc_info: Optional[Exception] = None,
        extra: Optional[Dict[str, Any]] = None,
        stack_info: bool = False,
    ) -> Context:
        return self.log(
            logging.ERROR,
            msg,
            exc_info=exc_info,
            extra=extra,
            stack_info=stack_info,
            stacklevel=2,
            *args,
        )

    def critical(
        self,
        msg: str,
        *args: Any,
        exc_info: Optional[Exception] = None,
        extra: Optional[Dict[str, Any]] = None,
        stack_info: bool = False,
    ) -> Context:
        return self.log(
            logging.CRITICAL,
            msg,
            exc_info=exc_info,
            extra=extra,
            stack_info=stack_info,
            stacklevel=2,
            *args,
        )

    def success(
        self,
        msg: str,
        *args: Any,
        exc_info: Optional[Exception] = None,
        extra: Optional[Dict[str, Any]] = None,
        stack_info: bool = False,
    ) -> Context:
        return self.log(
            SUCCESS,
            msg,
            exc_info=exc_info,
            extra=extra,
            stack_info=stack_info,
            stacklevel=2,
            *args,
        )


class ContextLogger(logging.Logger):
    def with_field(self, **kwargs: Any) -> Context:
        return Context(self, context=kwargs)

    def success(
        self,
        msg: str,
        exc_info: Optional[Exception] = None,
        extra: Optional[Dict[str, Any]] = None,
        stack_info: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        return self._log(  # type: ignore
            SUCCESS,
            msg,
            args,
            exc_info=exc_info,
            extra=extra,
            stack_info=stack_info,
            **kwargs,
        )



import json
from typing import Any, Dict, List

from colorama import Fore, Style

level_colors: Dict[str, str] = {
    "CRITICAL": Fore.RED,
    "ERROR": Fore.MAGENTA,
    "WARNING": Fore.YELLOW,
    "INFO": Fore.LIGHTCYAN_EX,
    "DEBUG": Fore.LIGHTBLACK_EX,
    "SUCCESS": Fore.GREEN,
}


def dumps(record: Dict[str, Any], *args: Any, **kwargs: Any) -> str:
    level = record.pop("levelname", "UNSET")
    color = record.pop("color", level_colors.get(level, ""))
    filename = record.pop("filename", "")
    lineno = record.pop("lineno", "")
    parts: List[str] = []
    for k, v in record.items():
        parts.append(
            "%s%s: %s%s" % (color, k, Style.RESET_ALL, json.dumps(v, *args, **kwargs))
        )
    return "%s[%s]%s(%s:%s) %s" % (
        color,
        level,
        Style.RESET_ALL,
        filename,
        lineno,
        "; ".join(parts),
    )



from fastapi import Request

from ..exceptions import HTTPUnAuthorizedError


def gen():
    async def get_current_user(request: Request) -> str:
        user_id: str = request.session.get("user_id")  # type: ignore
        if not user_id:
            raise HTTPUnAuthorizedError
        return user_id

    def authorize(request: Request, user_id: str):
        request.session["user_id"] = user_id  # type: ignore

    return get_current_user, authorize



from typing import Any, Callable, Coroutine, Union

from fastapi import Depends
from passlib.context import CryptContext

from ..g import set_user

password_context: CryptContext = CryptContext(schemes=["bcrypt"])


def with_context(get_current_user: Callable[..., Union[Coroutine[Any, Any, Any], Any]]):
    async def f(current_user: Any = Depends(get_current_user)):
        set_user(current_user)
        return current_user

    return f



import asyncio
from functools import wraps
from typing import Any, Callable


def coro(f: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any):
        return asyncio.run(f(*args, **kwargs))

    return wrapper



from typing import Any, Dict

from tortoise import Tortoise

from .base import BaseModel
from .enum import IntEnum, StrEnum
from .recoverialbe import RecoverableModel


async def init(config: Dict[str, Any]):
    await Tortoise.init(config=config)  # type: ignore


def init_models(config: Dict[str, Any]):
    for name, info in config["apps"].items():
        Tortoise.init_models(info["models"], name)


__all__ = ["StrEnum", "IntEnum", "BaseModel", "RecoverableModel"]



import time

import pytest
from faker import Faker
from jose import ExpiredSignatureError

from anyforce.api.security import jwt, password_context

from .model import User


@pytest.mark.asyncio
async def test_jwt(database: bool, faker: Faker):
    assert database
    email = faker.email()
    password = faker.pystr()
    await User.create(
        email=email,
        hashed_password=password_context.hash(password),
    )
    get_current_user, authorize = jwt.gen("", faker.name(), expire_after_seconds=1)
    token = authorize(email)
    authed_user = await get_current_user(token)
    assert authed_user == email

    time.sleep(2)
    try:
        await get_current_user(token)
        assert False
    except ExpiredSignatureError:
        pass
